#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <mmsystem.h>

/*
cl runtime.c /link winmm.lib
runtime prog [args]

cl /DRDTSC runtime.c rdtsc.obj /link winmm.lib
runtimeRDTSC prog [args]

cl /DQPC runtime.c /link winmm.lib
runtimeQPC prog [args]
*/

#ifdef RDTSC
extern ULONGLONG rdtsc(void);
#endif

void lasterr(DWORD lasterror,FILE *errout,char *fmt, ...)
{
    LPVOID lpMsgBuf;
    va_list argp;

    if ( fmt != NULL && *fmt != 0 ) {
        va_start(argp, fmt);
        vfprintf(errout,fmt,argp);
        va_end(argp);
    }
    FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        lasterror,
        MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0,
        NULL
    );
    fprintf(errout,"%s",lpMsgBuf);
    LocalFree(lpMsgBuf);
}

int main(int argc,char **argv)
{
    char *p;
    int ec;
    SECURITY_ATTRIBUTES ProcessAttributes;
    SECURITY_ATTRIBUTES ThreadAttributes;
    STARTUPINFO StartupInfo;
    PROCESS_INFORMATION ProcessInformation;
    DWORD ExitCode;
    BOOL ret;
#ifdef RDTSC
    ULONGLONG tsc;
    ULONGLONG tsc1;
    ULONGLONG tsc2;
#endif
#ifdef QPC
    LARGE_INTEGER tq1;
    LARGE_INTEGER tq2;
    LARGE_INTEGER frq;
    BOOL retq1;
    BOOL retq2;
    int qpcerr1;
    int qpcerr2;
#endif
    DWORD t1;
    DWORD t2;
    DWORD t3;
    DWORD t4;
    DWORD ts1;
    DWORD ts2;
    DWORD lasterror;
    DWORD_PTR affinitymask;
    MMRESULT timeperiod;

    if ( argc < 2 ) {
#ifdef RDTSC
        printf("runtimeRDTSC prog [args]\n");
#else
#ifdef QPC
        printf("runtimeQPC prog [args]\n");
#else
        printf("runtime prog [args]\n");
#endif
#endif
        return 0;
    }
    p = strdup(GetCommandLine());
    while ( *p == ' ' ) p++;
    if ( *p == 0 ) {
        ;
    } else if ( *p == '"' ) {
        ec = '"';
        p++;
    } else {
        ec = ' ';
    }
    while ( *p ) {
        if ( *p == ec ) {
            if ( *p == ' ' ) {
                *p = 0;
                p++;
                break;
            } else {
                *p = 0;
                p++;
                break;
            }
        }
        p++;
    }
    while ( *p == ' ' ) p++;
    ProcessAttributes.nLength = sizeof ProcessAttributes;
    ProcessAttributes.lpSecurityDescriptor = NULL;
    ProcessAttributes.bInheritHandle = TRUE;
    ThreadAttributes.nLength = sizeof ThreadAttributes;
    ThreadAttributes.lpSecurityDescriptor = NULL;
    ThreadAttributes.bInheritHandle = TRUE;
    GetStartupInfo(&StartupInfo);
/*
    StartupInfo.cb = sizeof StartupInfo;
    StartupInfo.lpReserved = NULL;
*/
    ExitCode = 0;
#if defined(RDTSC) || defined(QPC)
    affinitymask = SetThreadAffinityMask(GetCurrentThread(),1);
    if ( affinitymask == 0 ) {
        fprintf(stderr,"SetThreadAffinityMask() failed (%d)\n",GetLastError());
    }
#endif
    timeperiod = timeBeginPeriod(1);
    if ( timeperiod != TIMERR_NOERROR ) {
        fprintf(stderr,"timeBeginPeriod(1) failed (%d)\n",timeperiod);
    }
    t1 = GetTickCount();
    ts1 = timeGetTime();
#ifdef RDTSC
    tsc1 = rdtsc();
#endif
#ifdef QPC
    tq1.QuadPart = 0;
    tq2.QuadPart = 0;
    retq1 = QueryPerformanceCounter(&tq1);
    if ( retq1 == 0 ) {
        qpcerr1 = GetLastError();
    }
#endif
    ret = CreateProcess(NULL,// s\W[̖O
      p,                     // R}hC̕
      &ProcessAttributes,    // ZLeBLqq
      &ThreadAttributes,     // ZLeBLqq
      TRUE,                  // nȟpIvV
      CREATE_DEFAULT_ERROR_MODE | NORMAL_PRIORITY_CLASS,// 쐬̃tO
      NULL,                  // VubN
      NULL,                  // JgfBNg̖O
      &StartupInfo,          // X^[gAbv
      &ProcessInformation    // vZX
      );
    if ( ret == 0 ) {
        lasterror = GetLastError();
        lasterr(lasterror,stderr,"CreateProcess() failed %d ",lasterror);
        return 255;
    } else {
        ret = WaitForSingleObject(ProcessInformation.hProcess, INFINITE);
#ifdef RDTSC
        tsc2 = rdtsc();
#endif
#ifdef QPC
        retq2 = QueryPerformanceCounter(&tq2);
        if ( retq2 == 0 ) {
            qpcerr2 = GetLastError();
        }
#endif
        ts2 = timeGetTime();
        t2 = GetTickCount();
        if ( ret == WAIT_FAILED ) {
            fprintf(stderr,"WAIT_FAILED %d\n",GetLastError());
        } else if ( ret == WAIT_ABANDONED ) {
            fprintf(stderr,"WAIT_ABANDONED\n");
        }
        ret = GetExitCodeProcess(ProcessInformation.hProcess,&ExitCode); 
        if ( ret == 0 ) {
            fprintf(stderr,"GetExitCodeProcess() failed %d\n",GetLastError());
        } else {
            if ( ExitCode == STILL_ACTIVE ) {
                fprintf(stderr,"ERRORLEVEL=%u (STILL_ACTIVE)\n",ExitCode);
            } else {
                fprintf(stderr,"ERRORLEVEL=%u\n",ExitCode);
            }
        }
        CloseHandle(ProcessInformation.hThread) ;
        CloseHandle(ProcessInformation.hProcess) ;
#ifdef RDTSC
        fprintf(stderr,"tsc1=%I64u tsc2=%I64u tsc2-tsc1=%I64u\n",tsc1,tsc2,tsc = tsc2 - tsc1);
        tsc1 = rdtsc();
        t3 = GetTickCount();
        while ( (t4 = GetTickCount() - t3) < 1000 ) ;
        tsc2 = rdtsc();
/*        fprintf(stderr,"t4=%u\n",t4);*/
        fprintf(stderr,"cpu speed:%I64u\n",(tsc2 - tsc1) * t4 / 1000);
        fprintf(stderr,"time:%f\n",(__int64)tsc * 1000000.0 / (__int64)(tsc2 - tsc1) / t4);
#endif
#ifdef QPC
        if ( retq1 == 0 ) {
            fprintf(stderr,"QueryPerformanceCounter(tq1) failed (%d)\n",qpcerr1);
        }
        if ( retq2 == 0 ) {
            fprintf(stderr,"QueryPerformanceCounter(tq2) failed (%d)\n",qpcerr2);
        }
        fprintf(stderr,"tq1=%I64u tq2=%I64u tq2-tq1=%I64u\n",tq1.QuadPart,tq2.QuadPart,tq2.QuadPart - tq1.QuadPart);
        frq.QuadPart = 0;
        ret = QueryPerformanceFrequency(&frq);
        if ( ret == 0 ) {
            fprintf(stderr,"QueryPerformanceFrequency() failed (%d)\n",GetLastError());
        }
        fprintf(stderr,"frequency=%I64u\n",frq.QuadPart);
        fprintf(stderr,"time=%f\n",(tq2.QuadPart - tq1.QuadPart) * 1000.0 / frq.QuadPart);
#if 0
        tq1.QuadPart = 0;
        tq2.QuadPart = 0;
        retq1 = QueryPerformanceCounter(&tq1);
        if ( retq1 == 0 ) {
            qpcerr1 = GetLastError();
        }
        t3 = GetTickCount();
        while ( (t4 = GetTickCount() - t3) < 1000 ) ;
        retq2 = QueryPerformanceCounter(&tq2);
        if ( retq1 == 0 ) {
            fprintf(stderr,"QueryPerformanceCounter(tq1) failed (%d)\n",qpcerr1);
        }
        if ( retq2 == 0 ) {
            qpcerr2 = GetLastError();
        }
        if ( retq2 == 0 ) {
            fprintf(stderr,"QueryPerformanceCounter(tq2) failed (%d)\n",qpcerr2);
        }
        fprintf(stderr,"cpu speed:%I64u\n",(tq2.QuadPart - tq1.QuadPart) * t4 / 1000);
#endif
#endif
        timeperiod = timeEndPeriod(1);
        if ( timeperiod != TIMERR_NOERROR ) {
            fprintf(stderr,"timeEndPeriod(1) failed (%d)\n",timeperiod);
        }
        fprintf(stderr,"t1=%u t2=%u t2-t1=%u\n",t1,t2,t2-t1);
        fprintf(stderr,"ts1=%u ts2=%u ts2-ts1=%u\n",ts1,ts2,ts2-ts1);
        return ExitCode;
    }
}
